import queue


class MessageAnnouncer:
    def __init__(self):
        self.listeners = []

    def listen(self):
        self.listeners.append(queue.Queue(maxsize=10))
        return self.listeners[-1]

    def announce(self, msg):
        for i in reversed(range(len(self.listeners))):
            try:
                self.listeners[i].put_nowait(msg)
            except queue.Full:
                del self.listeners[i]


METRICS_ANNOUNCER = MessageAnnouncer()
EVENTS_ANNOUNCER = MessageAnnouncer()



from django.apps import AppConfig


class DjangoRiverMLConfig(AppConfig):
    name = "django_river_ml"
    verbose_name = "River Online Machine Learning for Django"

    def ready(self):
        import django_river_ml.signals

        assert django_river_ml.signals



from django_river_ml import settings
import abc
import contextlib
import os
import shelve

import river.base
import river.metrics
import river.stats
import river.utils
import dill
import redis

import logging

logger = logging.getLogger(__name__)

from . import exceptions
from . import flavors
from . import namer


class StorageBackend(abc.ABC):
    """Abstract storage backend.

    This interface defines a set of methods to implement in order for a database to be used as a
    storage backend. This allows using different databases in a homogeneous manner by proving a
    single interface. Since online-ml models are largely defined by Python dictionaries, we use
    key value store databases like redis.

    """

    @abc.abstractmethod
    def __setitem__(self, key, obj):
        """Store an object."""

    @abc.abstractmethod
    def __getitem__(self, key):
        """Retrieve an object."""

    @abc.abstractmethod
    def __delitem__(self, key):
        """Remove an object from storage."""

    @abc.abstractmethod
    def __iter__(self):
        """Iterate over the keys."""

    @abc.abstractmethod
    def close(self):
        """Do something when the app shuts down."""

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default


class ShelveBackend(shelve.DbfilenameShelf, StorageBackend):  # type: ignore
    """Storage backend based on the shelve module from the standard library.

    This should mainly be used for development and testing, but not production.

    """


class RedisBackend(StorageBackend):
    """Redis is the suggest backend for a more production database."""

    def __init__(self, host, port, db):
        self.r = redis.Redis(host=host, port=port, db=db)

    def __setitem__(self, key, obj):
        self.r[key] = dill.dumps(obj)

    def __getitem__(self, key):
        return dill.loads(self.r[key])

    def __delitem__(self, key):
        self.r.delete(key)

    def __iter__(self):
        for key in self.r.scan_iter():
            yield key.decode()

    def close(self):
        return


# The following will make it so that shelve.open returns ShelveBackend instead of DbfilenameShelf
shelve.DbfilenameShelf = ShelveBackend  # type: ignore


def get_db() -> StorageBackend:
    """
    Get the database, an attribute of settings.
    """
    if not hasattr(settings, "db"):
        backend = settings.STORAGE_BACKEND

        if backend == "shelve":
            settings.db = shelve.open(settings.SHELVE_PATH)

        elif backend == "redis":
            settings.db = RedisBackend(
                host=settings.REDIS_HOST,
                port=int(settings.REDIS_PORT),
                db=int(settings.REDIS_DB),
            )
        else:
            raise ValueError(f"Unknown storage backend: {backend}")

    return settings.db


def close_db(e=None):
    if hasattr(settings, "db"):
        if settings.db is not None:
            settings.db.close()
        delattr(settings, "db")


def drop_db():
    """This function's responsability is to wipe out a database.

    This could be implement within each StorageBackend, it's just a bit more akward because at this
    point the database connection is not stored in the app anymore.
    """
    backend = settings.STORAGE_BACKEND

    if backend == "shelve":
        path = settings.SHELVE_PATH
        with contextlib.suppress(FileNotFoundError):
            os.remove(f"{path}")

    elif backend == "redis":
        r = redis.Redis(
            host=settings.REDIS_HOST,
            port=int(settings.REDIS_PORT or 6379),
            db=int(settings.REDIS_DB or 0),
        )
        r.flushdb()


def set_flavor(flavor: str, name: str):
    try:
        flavor = flavors.allowed_flavors()[flavor]
    except KeyError:
        raise exceptions.UnknownFlavor

    db = get_db()
    db[f"flavor/{name}"] = flavor


def init_stats(name: str):
    db = get_db()
    db[f"stats/{name}"] = {
        "learn_mean": river.stats.Mean(),
        "learn_ewm": river.stats.EWMean(0.3),
        "predict_mean": river.stats.Mean(),
        "predict_ewm": river.stats.EWMean(0.3),
    }


def init_metrics(name: str):
    db = get_db()
    try:
        flavor = db[f"flavor/{name}"]
    except KeyError:
        raise exceptions.FlavorNotSet
    db[f"metrics/{name}"] = flavor.default_metrics()


def add_model(model: river.base.Estimator, flavor: str, name: str = None) -> str:
    db = get_db()

    # Pick a name if none is given
    if name is None:
        while True:
            name = _random_slug()
            if f"models/{name}" not in db:
                break

    # Make sure flavor is valid before continuing
    # it will be associated with the model name
    set_flavor(flavor=flavor, name=name)
    db[f"models/{name}"] = model

    init_stats(name)
    init_metrics(name)
    return name


def delete_model(name: str):
    db = get_db()
    del db[f"models/{name}"]
    del db[f"stats/{name}"]
    del db[f"metrics/{name}"]


def _random_slug() -> str:
    return namer.namer.generate()



def get_server(request):
    """
    Given a request, parse it to determine the server name and using http/https
    """
    scheme = request.is_secure() and "https" or "http"
    return f"{scheme}://{request.get_host()}"


def format_sse(data: str, event=None) -> str:
    """
    >>> format_sse(data=json.dumps({'abc': 123}), event='Jackson 5')
    'event: Jackson 5\\ndata: {"abc": 123}\\n\\n'
    """
    msg = f"data: {data}\n\n"
    if event is not None:
        msg = f"event: {event}\n{msg}"
    return msg


def humanize_ns(ns: int) -> str:

    if ns == 0:
        return "0ns"

    μs = ("μs", 1000)
    ms = ("ms", μs[1] * 1000)
    s = ("s", ms[1] * 1000)
    m = ("m", s[1] * 60)

    rep = ""

    for d in (m, s, ms, μs):
        k, ns = divmod(ns, d[1])
        if k:
            rep += f"{k}{d[0]}"

    if ns:
        rep += f"{ns}ns"

    return rep



from django.core.management.utils import get_random_secret_key
from django.conf import settings
import logging
import uuid
import os

logger = logging.getLogger(__name__)

here = os.path.dirname(os.path.abspath(__file__))
root = os.path.dirname(here)

# Default views to Authenticate
authenticated_views = [
    "django_river_ml.views.learn.view",
    "django_river_ml.views.predict.view",
    "django_river_ml.views.model.view",
    "django_river_ml.views.metrics.view",
    "django_river_ml.views.metrics.stream_events",
    "django_river_ml.views.metrics.stream_metrics",
]


timed_views = ["django_river_ml.views.learn.view", "django_river_ml.views.predict.view"]

# TODO do we need this?
authenticated_prefixes = ["/api/model/"]

# Defaults for models and storage

backends = ["shelve", "redis"]


DEFAULTS = {
    # Url base prefix
    "URL_PREFIX": "api",
    # Model storage
    "STORAGE_BACKEND": "shelve",
    # Redis (if used)
    "REDIS_DB": "river-redis",
    "REDIS_HOST": "localhost",
    "REDIS_PORT": "6379",
    # Base directory for storing secrets
    "APP_DIR": root,
    "CACHE_DIR": None,
    # Disable authentication
    "DISABLE_AUTHENTICATION": True,
    # Domain used in templates, api prefix
    "DOMAIN_URL": "http://127.0.0.1:8000",
    # The number of seconds a session (upload request) is valid (10 minutes)
    "SESSION_EXPIRES_SECONDS": 600,
    # The number of seconds a token is valid (10 minutes)
    "TOKEN_EXPIRES_SECONDS": 600,
    # Disable deletion of an image by tag or manifest (default is not disabled)
    # Default views to put under authentication, given that DISABLE_AUTHENTICTION is False
    "AUTHENTICATED_VIEWS": authenticated_views,
    # If you have a custom authentication server to generate tokens (defaults to /registry/auth/token
    "AUTHENTICATION_SERVER": None,
    # jwt encoding secret: set server wide or generated on the fly
    "JWT_SERVER_SECRET": str(uuid.uuid4()),
    # View rate limit, defaults to 100/1day using django-ratelimit based on ipaddress
    "VIEW_RATE_LIMIT": "10000/1d",
    # Given that someone goes over, are they blocked for a period?
    "VIEW_RATE_LIMIT_BLOCK": True,
    # Shelve and jwt keys (will be generated if not found)
    "SHELVE_SECRET_KEY": None,
    "JWT_SECRET_KEY": None,
}

# The user can define a section for django_river_ml in settings
CACHES = getattr(settings, "CACHES", [])
MIDDLEWARE = getattr(settings, "MIDDLEWARE", [])

ml = getattr(settings, "django_river_ml", DEFAULTS)

# Add middleware to calculate times and stats
timer_middleware = "django_river_ml.middleware.timer_middleware"

if timer_middleware not in MIDDLEWARE:
    MIDDLEWARE.append(timer_middleware)

# Over-ride defaults with user settings!
for key, default in DEFAULTS.items():

    # If we find it in the environment, use this first
    value = os.environ.get(key)

    # Handle environment booleans
    if not value:
        value = ml.get(key, DEFAULTS[key])
    elif value.lower() == "true":
        value = True
    elif value == "false":
        value = False

    # Fall back to user settings and defaults
    locals()[key] = value

# Default shelve path is root of install - should be updated to be within app
SHELVE_PATH = ml.get("SHELVE_PATH", os.path.join(APP_DIR, "shelve.db"))

# Validation of django-river-ml settings
if STORAGE_BACKEND not in backends:
    sys.exit(
        "%s is not a valid storage backend, but me one of %s"
        % (STORAGE_BACKEND, " ".join(backends))
    )

# The database state will be held via settings
# db = None

# Generate secret key for shelve if used
def generate_secret_keys(filename):
    """
    A helper function to write a randomly generated secret key to file
    """
    with open(filename, "w") as fd:
        for keyname in ["SHELVE_SECRET_KEY", "JWT_SERVER_SECRET"]:
            key = get_random_secret_key()
            fd.writelines("%s = '%s'\n" % (keyname, key))


# **Important** this will generate in the install directory
# so recommended to already add to app settings.
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# Generate secret key for API (jwt) and shelve if do not exist, and not defined in environment
if not SHELVE_SECRET_KEY or not JWT_SERVER_SECRET:
    try:
        from .secret_key import SHELVE_SECRET_KEY, JWT_SERVER_SECRET
    except ImportError:
        generate_secret_keys(os.path.join(BASE_DIR, "secret_key.py"))
        from .secret_key import SHELVE_SECRET_KEY, JWT_SERVER_SECRET

# Create a filesystem cache for jwt tokens
cache = CACHE_DIR or os.path.join(APP_DIR, "cache")
if not os.path.exists(cache):
    logger.debug(f"Creating cache directory {cache}")
    os.makedirs(cache)

CACHES.update(
    {
        "django_river_ml": {
            "BACKEND": "django.core.cache.backends.filebased.FileBasedCache",
            "LOCATION": os.path.abspath(cache),
        }
    }
)



import river
from rest_framework.views import APIView
from rest_framework.response import Response
from ratelimit.decorators import ratelimit

from django_river_ml import settings
from django_river_ml.version import __version__
from django.utils.decorators import method_decorator


class ServiceInfo(APIView):
    """
    provide version support information based on response statuses.
    """

    permission_classes = []
    allowed_methods = ("GET",)

    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        print("GET /api/")

        data = {
            "id": "django_river_ml",
            "status": "running",
            "name": "Django River ML Endpoint",
            "description": "This service provides an api for models",
            "documentationUrl": "https://vsoch.github.io/django-river-ml",
            "storage": settings.STORAGE_BACKEND,
            "river_version": river.__version__,
            "version": __version__,
        }
        return Response(status=200, data=data)



from .auth import GetAuthToken
from .base import ServiceInfo
from .learn import LearnView
from .metrics import MetricsView, stream_metrics, stream_events, StatsView
from .model import ModelView, ModelsView, ModelDownloadView
from .predict import PredictView



from django.urls import resolve
from django.contrib.auth.models import User

from django_river_ml import settings
import django_river_ml.utils as utils

from rest_framework.authtoken.models import Token
from rest_framework.response import Response

from django.middleware import cache

from datetime import datetime
import uuid
import base64
import re
import time
import jwt


def is_authenticated(request):
    """
    Function to check if a request is authenticated.
    Returns a boolean to indicate if the user is authenticated, and a response with
    the challenge if not.

    request (requests.Request)    : the Request object to inspect
    """
    # Derive the view name from the request PATH_INFO
    func, _, _ = resolve(request.META["PATH_INFO"])
    view_name = "%s.%s" % (func.__module__, func.__name__)

    print("VIEW NAME: %s" % view_name)
    # If authentication is disabled, return the original view
    if settings.DISABLE_AUTHENTICATION or view_name not in settings.AUTHENTICATED_VIEWS:
        return True, None, None

    # Case 2: Already has a jwt valid token
    is_valid, user = validate_jwt(request)
    if is_valid:
        return True, None, user

    # Case 3: False and response will return request for auth
    user = get_user(request)
    if not user:
        headers = {"Www-Authenticate": get_challenge(request)}
        return False, Response(status=401, headers=headers), user

    # Denied for any other reason
    return False, Response(status=403), user


def generate_jwt(username):
    """Given a username generate a jwt
    token to return to the user with a default expiration of 10 minutes.

    username (str)  : the user's username to add under "sub"
    """
    # The jti expires after TOKEN_EXPIRES_SECONDS
    issued_at = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
    filecache = cache.caches["django_river_ml"]
    jti = str(uuid.uuid4())
    filecache.set(jti, "good", timeout=settings.TOKEN_EXPIRES_SECONDS)
    now = int(time.time())
    expires_at = now + settings.TOKEN_EXPIRES_SECONDS

    # import jwt and generate token
    # https://tools.ietf.org/html/rfc7519#section-4.1.5
    payload = {
        "sub": username,
        "exp": expires_at,
        "nbf": now,
        "iat": now,
        "jti": jti,
    }
    token = jwt.encode(payload, settings.JWT_SERVER_SECRET, algorithm="HS256")
    if isinstance(token, bytes):
        token = token.decode("utf-8")
    return {
        "token": token,
        "expires_in": settings.TOKEN_EXPIRES_SECONDS,
        "issued_at": issued_at,
    }


def validate_jwt(request):
    """
    Given a jwt token, decode and validate

    request (requests.Request)    : the Request object to inspect
    """
    header = request.META.get("HTTP_AUTHORIZATION", "")
    if re.search("bearer", header, re.IGNORECASE):
        encoded = re.sub("bearer", "", header, flags=re.IGNORECASE).strip()

        # Any reason not valid will issue an error here
        try:
            decoded = jwt.decode(
                encoded, settings.JWT_SERVER_SECRET, algorithms=["HS256"]
            )
        except Exception as exc:
            print("jwt could no be decoded, %s" % exc)
            return False, None

        # Ensure that the jti is still valid
        filecache = cache.caches["django_river_ml"]
        if not filecache.get(decoded.get("jti")) == "good":
            print("Filecache with jti not found.")
            return False, None

        # The user must exist
        try:
            user = User.objects.get(username=decoded.get("sub"))
            return True, user
        except User.DoesNotExist:
            print("Username %s not found" % decoded.get("sub"))
            return False, None

    return False, None


def get_user(request):
    """Given a request, read the Authorization header to get the base64 encoded
    username and token (password) which is a basic auth. If we return the user
    object, the user is successfully authenticated. Otherwise, return None.
    and the calling function should return Forbidden status.

    request (requests.Request)    : the Request object to inspect
    """
    header = request.META.get("HTTP_AUTHORIZATION", "")
    if re.search("basic", header, re.IGNORECASE):
        encoded = re.sub("basic", "", header, flags=re.IGNORECASE).strip()
        decoded = base64.b64decode(encoded).decode("utf-8")
        username, token = decoded.split(":", 1)
        try:
            token = Token.objects.get(key=token)
            if token.user.username == username:
                return token.user
        except:
            pass


def get_token(request):
    """The same as validate_token, but return the token object to check the
    associated user.

    request (requests.Request)    : the Request object to inspect
    """
    # Coming from HTTP, look for authorization as bearer token
    token = request.META.get("HTTP_AUTHORIZATION")

    if token:
        try:
            return Token.objects.get(key=token.replace("BEARER", "").strip())
        except Token.DoesNotExist:
            pass

    # Next attempt - try to get token via user session
    elif request.user.is_authenticated and not request.user.is_anonymous:
        try:
            return Token.objects.get(user=request.user)
        except Token.DoesNotExist:
            pass


def get_challenge(request):
    """Given an unauthenticated request, return a challenge in
    the Www-Authenticate header

    request (requests.Request): the Request object to inspect
    """
    DOMAIN_NAME = utils.get_server(request)
    auth_server = "%s/auth/token" % DOMAIN_NAME
    return 'realm="%s",service="%s"' % (
        auth_server,
        DOMAIN_NAME,
    )



from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.dispatch import receiver
from rest_framework.authtoken.models import Token

import sys

UserModel = get_user_model()


def get_user_token(user):
    """
    Get a user token.
    """
    try:
        return str(Token.objects.get(user=user))
    except Token.DoesNotExist:
        sys.exit("Token for %s does not exist." % user)


def create_user_token(user):
    """
    Function to create the token for the user, if it doesn't exist.
    """
    try:
        token = Token.objects.get(user=user)
    except Token.DoesNotExist:
        token = Token.objects.create(user=user)
    return str(token)


@receiver(post_save, sender=UserModel)
def create_auth_token(sender, instance=None, created=False, **kwargs):
    """Create a token for the user when the user is created

    1. Assign user a token

    Create a token instance for all newly created User instances. We only
    run on user creation to avoid having to check for existence on each call
    to User.save. We also check if one is already created in case another app
    or the main user application is running the same function.
    """
    if created:
        create_user_token(instance)



from river.metrics.base import ClassificationMetric

import django_river_ml.storage as storage
import django_river_ml.exceptions as exceptions
import django_river_ml.utils as utils
import django_river_ml.announce as announce
import django_river_ml.flavors as flavors

import json
import copy


class RiverClient:
    """
    A river client includes shared functions for interacting with storage.
    This implementation was chosen so it can easily be plugged into another
    kind of server without needing to deal with the server-specific requests.
    """

    def __init__(self):
        self.db = storage.get_db()

    def stats(self, model_name):
        try:
            stats = self.db[f"stats/{model_name}"]
        except KeyError:
            raise exceptions.InvalidUsage(message="We don't have stats for that model")

        return {
            "predict": {
                "n_calls": int(stats["predict_mean"].n),
                "mean_duration": int(stats["predict_mean"].get()),
                "mean_duration_human": utils.humanize_ns(
                    int(stats["predict_mean"].get())
                ),
                "ewm_duration": int(stats["predict_ewm"].get()),
                "ewm_duration_human": utils.humanize_ns(
                    int(stats["predict_ewm"].get())
                ),
            },
            "learn": {
                "n_calls": int(stats["learn_mean"].n),
                "mean_duration": int(stats["learn_mean"].get()),
                "mean_duration_human": utils.humanize_ns(
                    int(stats["learn_mean"].get())
                ),
                "ewm_duration": int(stats["learn_ewm"].get()),
                "ewm_duration_human": utils.humanize_ns(int(stats["learn_ewm"].get())),
            },
        }

    def stream_events(self):
        """
        Stream events from the events announcer.
        """
        messages = announce.EVENTS_ANNOUNCER.listen()
        return self._stream_announcer(messages)

    def stream_metrics(self):
        """
        Stream events from the metrics announcer.
        """
        messages = announce.METRICS_ANNOUNCER.listen()
        return self._stream_announcer(messages)

    def metrics(self):
        """
        Get metrics from the database
        """
        try:
            metrics = self.db["metrics"]
        except KeyError:
            raise exceptions.FlavorNotSet
        return {metric.__class__.__name__: metric.get() for metric in metrics}

    def add_model(self, model, flavor, name=None):
        """
        Add a model by name (e.g., via API post)
        """
        # Validate the model
        ok, error = flavors.check(model, flavor)
        if not ok:
            return False, {"message": error}
        name = storage.add_model(model, name=name, flavor=flavor)
        return True, {"name": name}

    def get_model(self, name):
        """
        Get a model by name
        """
        return self.db[f"models/{name}"]

    def delete_model(self, name):
        """
        Delete a model by name.
        """
        key = f"models/{name}"
        if key not in self.db:
            return False
        del self.db[key]
        return True

    def models(self):
        """
        Get models known to a database.
        """
        model_names = sorted(
            [k.split("/", 1)[1] for k in self.db if k.startswith("models/")]
        )
        return {"models": model_names}

    def _stream_announcer(self, announcer):
        """
        Shared function to stream events from an announcer.
        """
        while True:
            item = announcer.get()  # blocks until a new message arrives
            yield item

    # Learn and predict!

    def learn(
        self,
        model_name,
        ground_truth=None,
        prediction=None,
        features=None,
        identifier=None,
    ):
        """
        A learning event takes a learning schema
        """
        # If an ID is given, then retrieve the stored info.
        try:
            memory = self.db["#%s" % identifier] if identifier else {}
        except KeyError:
            return False, f"No information stored for ID '{identifier}'"

        model_name = memory.get("model", model_name)
        features = memory.get("features", features)
        prediction = memory.get("prediction", prediction)

        # Raise an error if no features are provided
        if features is None:
            return False, "No features are stored and none were provided."

        try:
            model = self.db[f"models/{model_name}"]
        except KeyError:
            return False, f"No model named '{model_name}'."

        # Obtain a prediction if none was made earlier
        if prediction is None:
            flavor = self.db[f"flavor/{model_name}"]
            pred_func = getattr(model, flavor.pred_func)
            try:
                prediction = pred_func(x=copy.deepcopy(features))
            except Exception as e:
                return False, repr(e)

        metrics = self.update_metrics(prediction, ground_truth, model_name)

        # Update the model
        try:
            model.learn_one(x=copy.deepcopy(features), y=ground_truth)
        except Exception as e:
            return False, repr(e)

        self.db[f"models/{model_name}"] = model
        self.announce_event(
            "learn",
            {
                "model": model_name,
                "features": features,
                "prediction": prediction,
                "ground_truth": ground_truth,
            },
        )
        self.announce_metrics(metrics)

        # Delete the id from the db
        if identifier:
            self.delete_id(identifier)
        return True, "Successful learn."

    def predict(self, features, model_name, identifier=None, model=None):
        """
        Run a prediction
        """
        try:
            model = self.db[f"models/{model_name}"]
        except KeyError:
            return False, f"No model named '{model_name}'."

        # We make a copy because the model might modify the features in-place while we want to be able
        # to store an identical copy
        features = copy.deepcopy(features)

        # Make the prediction
        flavor = self.db[f"flavor/{model_name}"]
        pred_func = getattr(model, flavor.pred_func)
        try:
            pred = pred_func(x=features)
        except Exception as e:
            return False, repr(e)

        # The unsupervised parts of the model might be updated after a prediction, so we need to store it
        self.db[f"models/{model_name}"] = model

        # Announce the prediction
        self.announce_event(
            "predict",
            {
                "model": model_name,
                "features": features,
                "prediction": pred,
            },
        )

        # If an ID is provided, then we store the features in order to be able to use them for learning
        # further down the line. We remove granularity about creation (200 vs 201) here.
        created = False
        if identifier:
            self.db["#%s" % identifier] = {
                "model": model_name,
                "features": features,
                "prediction": pred,
            }
            created = True
        return True, {"model": model_name, "prediction": pred, "created": created}

    def announce_metrics(self, metrics):
        """
        Announce the current metric values
        """
        if announce.METRICS_ANNOUNCER.listeners:
            msg = json.dumps(
                {metric.__class__.__name__: metric.get() for metric in metrics}
            )
            announce.METRICS_ANNOUNCER.announce(utils.format_sse(data=msg))

    def announce_event(self, event, data):
        """
        Announce the event
        """
        if announce.EVENTS_ANNOUNCER.listeners:
            announce.EVENTS_ANNOUNCER.announce(
                utils.format_sse(data=json.dumps(data), event=event)
            )

    def update_metrics(self, prediction, ground_truth, model_name):
        """
        Given a prediction, update metrics to reflect it.
        """
        # Update the metrics
        metrics = self.db[f"metrics/{model_name}"]

        # At this point prediction is a dict. It might be empty because no training data has been seen
        if not prediction:
            return metrics

        for metric in metrics:

            # If the metrics requires labels but the prediction is a dict, then we need to retrieve the
            # predicted label with the highest probability
            if (
                isinstance(metric, ClassificationMetric)
                and metric.requires_labels
                and isinstance(prediction, dict)
            ):
                pred = max(prediction, key=prediction.get)
                metric.update(y_true=ground_truth, y_pred=pred)

            else:
                metric.update(y_true=ground_truth, y_pred=prediction)
        self.db[f"metrics/{model_name}"] = metrics
        return metrics

    def delete_id(self, identifier):
        try:
            del self.db["#%s" % identifier]
        except KeyError:
            pass



from rest_framework.views import APIView
from rest_framework.response import Response
from django.views.decorators.cache import never_cache
from django.http import StreamingHttpResponse

from ratelimit.decorators import ratelimit
from django.utils.decorators import method_decorator
from django.views.decorators.http import require_http_methods
from django_river_ml.client import RiverClient
import django_river_ml.settings as settings


class MetricsView(APIView):
    """
    Get metrics for a database
    """

    permission_classes = []
    allowed_methods = ("GET",)

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """
        GET /api/metrics/
        """
        client = RiverClient()
        return Response(status=200, data=client.metrics())


@never_cache
@require_http_methods(["GET"])
@ratelimit(
    key="ip", block=settings.VIEW_RATE_LIMIT_BLOCK, rate=settings.VIEW_RATE_LIMIT
)
def stream_metrics(request):
    """GET /api/stream/metrics/"""
    client = RiverClient()
    return StreamingHttpResponse(
        client.stream_metrics(), content_type="text/event-stream"
    )


@never_cache
@require_http_methods(["GET"])
@ratelimit(
    key="ip", block=settings.VIEW_RATE_LIMIT_BLOCK, rate=settings.VIEW_RATE_LIMIT
)
def stream_events(request):
    """
    GET /api/stream/events/
    """
    client = RiverClient()
    return StreamingHttpResponse(
        client.stream_events(), content_type="text/event-stream"
    )


class StatsView(APIView):
    """
    Get stats for the database
    """

    permission_classes = []
    allowed_methods = ("GET",)

    @never_cache
    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """GET /api/stats/"""

        client = RiverClient()
        return Response(status=200, data=client.stats())



import inspect


def model_to_dict(model):
    """
    Dump a model to json
    """

    def unwrap(params):
        result = {}
        for k, v in params.items():

            # Dict gets unwrapped
            if isinstance(v, dict):
                result[k] = unwrap(v)

            # List or tuple gets unwrapped
            elif isinstance(v, (list, tuple)):
                items = []
                for item in v:
                    if inspect.isclass(item):
                        items.append(item.__name__)
                    elif isinstance(item, dict):
                        unwrapped = unwrap(item)
                        if unwrapped:
                            items.append(unwrapped)
                    else:
                        items.append(str(item))
                result[k] = items

            # Class we derive name
            elif inspect.isclass(k):
                result[v] = k.__name__
            else:
                result[k] = v
        return result

    params = model._get_params()
    return unwrap(params)



import abc
import typing

from river import metrics

# A "flavor" is a different kind of online-ml model.
# We need to know the correct check / prediction to do for different kinds!
# https://github.com/online-ml/chantilly/blob/master/chantilly/flavors.py#L1


def check(model, flavor_name):
    """
    Check a model against all flavors available
    """
    for flavor in [RegressionFlavor, BinaryFlavor, MultiClassFlavor]:
        flavor = flavor()
        if flavor_name == flavor.name:
            return flavor.check_model(model)
    return False, "This model flavor %s is not recognized" % flavor_name


def allowed_flavors():
    return {f().name: f() for f in [RegressionFlavor, BinaryFlavor, MultiClassFlavor]}


class Flavor(abc.ABC):
    @abc.abstractproperty
    def name(self):
        pass

    @abc.abstractmethod
    def check_model(self, model: typing.Any) -> typing.Tuple[bool, str]:
        """Checks whether or not a model works for a flavor."""

    @abc.abstractmethod
    def default_metrics(self) -> typing.List[metrics.Metric]:
        """Default metrics to record globally as well as for each model."""

    @abc.abstractproperty
    def pred_func(self) -> str:
        """The name of the required prediction function."""


class RegressionFlavor(Flavor):
    @property
    def name(self):
        return "regression"

    def check_model(self, model):
        for method in ("learn_one", "predict_one"):
            if not hasattr(model, method):
                return False, f"The model does not implement {method}."
        return True, None

    def default_metrics(self):
        return [metrics.MAE(), metrics.RMSE(), metrics.SMAPE()]

    @property
    def pred_func(self):
        return "predict_one"


class BinaryFlavor(Flavor):
    @property
    def name(self):
        return "binary"

    def check_model(self, model):
        for method in ("learn_one", "predict_proba_one"):
            if not hasattr(model, method):
                return False, f"The model does not implement {method}."
        return True, None

    def default_metrics(self):
        return [
            metrics.Accuracy(),
            metrics.LogLoss(),
            metrics.Precision(),
            metrics.Recall(),
            metrics.F1(),
        ]

    @property
    def pred_func(self):
        return "predict_proba_one"


class MultiClassFlavor(Flavor):
    @property
    def name(self):
        return "multiclass"

    def check_model(self, model):
        for method in ("learn_one", "predict_proba_one"):
            if not hasattr(model, method):
                return False, f"The model does not implement {method}."
        return True, None

    def default_metrics(self):
        return [
            metrics.Accuracy(),
            metrics.CrossEntropy(),
            metrics.MacroPrecision(),
            metrics.MacroRecall(),
            metrics.MacroF1(),
            metrics.MicroPrecision(),
            metrics.MicroRecall(),
            metrics.MicroF1(),
        ]

    @property
    def pred_func(self):
        return "predict_proba_one"



from django.core.management.base import BaseCommand, CommandError
from django.contrib.auth import authenticate, get_user_model
from getpass import getpass
from django_river_ml.signals import get_user_token


User = get_user_model()


class Command(BaseCommand):
    """
    get a token for a user (typically to use the API)
    """

    def add_arguments(self, parser):
        parser.add_argument("username", nargs=1, default=None, type=str)

    help = "Get a user token to interact with the API"

    def handle(self, *args, **options):
        if not options["username"]:
            raise CommandError("Please provide a username")
        username = options["username"][0]
        print("Username: %s" % username)

        # The username must exist
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            raise CommandError("This username does not exist.")

        # Create the user and ask for password
        password = getpass("Enter Password:")
        user = authenticate(None, username=user.username, password=password)

        if user is not None:
            print(get_user_token(user))
        else:
            print("This password is incorrect.")



import time
from django.urls import resolve

from django_river_ml import settings as settings
import django_river_ml.storage as storage

from copy import deepcopy
import json

# The following are pre and post events that should be run alongside views
# to update either metrics or stats


def timer_middleware(get_response):
    """
    A custom class to start and end a counter.
    This should be used to decorate api.learn and api.predict
    """
    # One-time configuration and initialization.

    def middleware(request):

        # Derive the view name from the request PATH_INFO
        func, _, _ = resolve(request.META["PATH_INFO"])
        view_name = "%s.%s" % (func.__module__, func.__name__)
        in_timed_view = view_name in settings.timed_views

        # Grab the started at time
        started_at = time.perf_counter_ns()

        # Deep copy the request body to see if we hav a name param
        model_name = None
        try:
            copied = deepcopy(request.body)
            payload = json.loads(copied.decode("utf-8"))
            model_name = payload.get("model")
        except:
            pass

        response = get_response(request)

        # Cut out early if not in a timed view
        if not in_timed_view or not model_name:
            return response

        # Calculate the duration and add to stats
        duration = time.perf_counter_ns() - started_at
        db = storage.get_db()
        if f"stats/{model_name}" not in db:
            storage.init_stats(model_name)
        stats = db[f"stats/{model_name}"]

        if request.path == "/%s/learn/" % settings.URL_PREFIX:
            stats["learn_mean"].update(duration)
            stats["learn_ewm"].update(duration)
        elif request.path == "/%s/predict/" % settings.URL_PREFIX:
            stats["predict_mean"].update(duration)
            stats["predict_ewm"].update(duration)

        db[f"stats/{model_name}"] = stats
        return response

    return middleware



from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.decorators import authentication_classes, permission_classes
from django.http import HttpResponseForbidden

import django_river_ml.auth as auth


@authentication_classes([])
@permission_classes([])
class GetAuthToken(APIView):
    """
    Given a GET request for a token, validate and return it.
    """

    permission_classes = []
    allowed_methods = ("GET",)

    def get(self, request, *args, **kwargs):
        """GET /auth/token"""
        print("GET /auth/token")
        user = auth.get_user(request)

        # No token provided matching a user, no go
        if not user:
            return HttpResponseForbidden()

        # Generate the token data, a dict with token, expires_in, and issued_at
        data = auth.generate_jwt(user.username)
        return Response(status=200, data=data)



from rest_framework.views import APIView
from rest_framework.response import Response

from django.http import QueryDict, HttpResponse

from django_river_ml import settings as settings
from django_river_ml.client import RiverClient
from django_river_ml import model as models
from django_river_ml.auth import is_authenticated

from ratelimit.decorators import ratelimit
from django.utils.decorators import method_decorator
import dill


class ModelDownloadView(APIView):
    permission_classes = []
    allowed_methods = ("GET",)

    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """
        GET /api/model/download/<name>/
        """
        allow_continue, response, _ = is_authenticated(request)
        if not allow_continue:
            return response

        name = kwargs.get("name")
        if not name:
            return Response(status=400, data={"message": "A model name is required"})
        client = RiverClient()
        model = client.get_model(name)
        return HttpResponse(dill.dumps(model), content_type="application/octet-stream")


class ModelView(APIView):
    permission_classes = []
    allowed_methods = ("GET", "POST", "DELETE")

    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """
        GET /api/model/<name>/
        """
        allow_continue, response, _ = is_authenticated(request)
        if not allow_continue:
            return response

        name = kwargs.get("name")
        if not name:
            return Response(status=400, data={"message": "A model name is required"})
        client = RiverClient()
        model = client.get_model(name)
        dumped = models.model_to_dict(model)
        return Response(status=200, data=dumped)

    def post(self, request, *args, **kwargs):
        """
        POST /api/model/<flavor>/<name>
        """
        allow_continue, response, _ = is_authenticated(request)
        if not allow_continue:
            return response

        name = None

        # if we only have one arg, we have flavor
        if len(kwargs) == 1:
            flavor = kwargs.get("name")

        # if we only have one arg, we have flavor
        elif len(kwargs) == 2:
            flavor = kwargs.get("flavor")
            name = kwargs.get("name")

        model = dill.loads(request.body)
        client = RiverClient()
        added, data = client.add_model(model, name=name, flavor=flavor)
        if added:
            return Response(status=201, data=data)
        return Response(status=400, data=data)

    def delete(self, request, *args, **kwargs):
        """
        DELETE /api/model/<name>/
        """
        params = QueryDict(request.body)
        name = params.get("name")

        allow_continue, response, _ = is_authenticated(request)
        if not allow_continue:
            return response

        client = RiverClient()
        if not client.delete_model(name):
            return Response(status=404)
        return Response(204)


class ModelsView(APIView):
    permission_classes = []
    allowed_methods = ("GET",)

    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="GET",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def get(self, request, *args, **kwargs):
        """
        GET /api/models/
        """
        allow_continue, response, _ = is_authenticated(request)
        if not allow_continue:
            return response

        client = RiverClient()
        return Response(status=200, data=client.models())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


version = get_version("django_river_ml", "version.py")

with open("README.md") as fd:
    readme = fd.read()

setup(
    name="django-river-ml",
    version=version,
    description="Online machine learning with river and Django.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Vanessa Sochat",
    author_email="vsoch@users.noreply.github.io",
    url="https://github.com/vsoch/django-river-ml",
    packages=["django_river_ml"],
    include_package_data=True,
    install_requires=[
        "django",
        "djangorestframework",
        "django-ratelimit",
        "pyjwt",
        "river",
        "redis",
        "dill",
        "cerebrus",
    ],
    license="Apache Software License 2.0",
    zip_safe=False,
    keywords="online-ml,river,chantilly",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.1",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
    ],
)



from django.urls import path
from django_river_ml import views
from django_river_ml import settings

app_name = "django_river_ml"

urlpatterns = [
    path(
        "auth/token/",
        views.GetAuthToken.as_view(),
        name="auth_token",
    ),
    path(
        "%s/" % settings.URL_PREFIX,
        views.ServiceInfo.as_view(),
        name="service_info",
    ),
    path(
        "%s/learn/" % settings.URL_PREFIX,
        views.LearnView.as_view(),
        name="learn",
    ),
    path(
        "%s/predict/" % settings.URL_PREFIX,
        views.PredictView.as_view(),
        name="predict",
    ),
    path(
        "%s/metrics/" % settings.URL_PREFIX,
        views.MetricsView.as_view(),
        name="metrics",
    ),
    path(
        "%s/stream/metrics/" % settings.URL_PREFIX,
        views.stream_metrics,
        name="stream_metrics",
    ),
    path(
        "%s/stream/events/" % settings.URL_PREFIX,
        views.stream_events,
        name="stream_events",
    ),
    path(
        "%s/stats/" % settings.URL_PREFIX,
        views.StatsView.as_view(),
        name="stats",
    ),
    path(
        "%s/model/" % settings.URL_PREFIX,
        views.ModelView.as_view(),
        name="model",
    ),
    path(
        "%s/model/<str:name>/" % settings.URL_PREFIX,
        views.ModelView.as_view(),
        name="model",
    ),
    path(
        "%s/model/<str:flavor>/<str:name>/" % settings.URL_PREFIX,
        views.ModelView.as_view(),
        name="model",
    ),
    path(
        "%s/models/" % settings.URL_PREFIX,
        views.ModelsView.as_view(),
        name="models",
    ),
    path(
        "%s/model/download/<str:name>/" % settings.URL_PREFIX,
        views.ModelDownloadView.as_view(),
        name="model_download",
    ),
]



from . import flavors


class InvalidUsage(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        super().__init__()
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv["message"] = self.message
        return rv


class UnknownFlavor(Exception):
    def __init__(self, *args, **kwargs):
        message = (
            "Allowed flavors are "
            + ", ".join([f"'{flavor}'" for flavor in flavors.allowed_flavors()])
            + "."
        )
        super().__init__(message, *args, **kwargs)


class FlavorNotSet(InvalidUsage):
    def __init__(self, *args, **kwargs):
        super().__init__(message="No flavor has been set.", *args, **kwargs)



from rest_framework.views import APIView
from rest_framework.response import Response

from django_river_ml import settings
from django_river_ml.auth import is_authenticated

from ratelimit.decorators import ratelimit
from django.utils.decorators import method_decorator

from django_river_ml.client import RiverClient
import json


class PredictView(APIView):
    """
    PredictView is a prediction API endpoint for a model.
    """

    permission_classes = []
    allowed_methods = ("POST",)

    @method_decorator(
        ratelimit(
            key="ip",
            rate=settings.VIEW_RATE_LIMIT,
            method="POST",
            block=settings.VIEW_RATE_LIMIT_BLOCK,
        )
    )
    def post(self, request, *args, **kwargs):
        """
        POST /api/predict/
        """
        try:
            payload = json.loads(request.body.decode("utf-8"))
        except:
            return Response(status=400)

        model_name = payload.get("model")
        identifier = payload.get("identifier")
        features = payload.get("features")

        allow_continue, response, _ = is_authenticated(request)
        if not allow_continue:
            return response

        if not model_name or not features:
            return Response(status=400, message="model and features are required.")

        client = RiverClient()
        success, data = client.predict(
            features=features, identifier=identifier, model_name=model_name
        )
        if not success:
            return Response(status=400, data=data)
        if data["created"]:
            return Response(status=201, data=data)
        return Response(status=200, data=data)
